package com.google.android.apps.work.dpcsupport;

import android.accounts.AccountManager;
import android.content.Context;
import android.content.Intent;
import android.content.pm.PackageManager;
import android.content.pm.ResolveInfo;
import android.content.res.Resources;
import android.content.res.TypedArray;
import android.content.res.XmlResourceParser;
import android.support.annotation.WorkerThread;
import android.util.AttributeSet;
import android.util.Log;
import android.util.TypedValue;
import android.util.Xml;
import java.io.IOException;
import java.lang.reflect.Field;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import org.xmlpull.v1.XmlPullParser;
import org.xmlpull.v1.XmlPullParserException;

/** Gets authenticator types present on the device. */
class AuthenticatorTypesGetter {

  /** Returns all the authenticator types mapped to the package name present on the device. */
  @WorkerThread
  public Map<String, String> getAuthenticatorTypeToPackageNameMap(Context context) {
    // The top choice of system API is AccountManager.getAuthenticatorTypes(). However during a
    // Intent.ACTION_PACKAGE_ADDED broadcast the returned list is not up to date (b/30979262).
    // Go the long way and parse the authenticator metadata from PackageManager.
    PackageManager pm = context.getPackageManager();
    Intent authenticatorIntent = new Intent(AccountManager.ACTION_AUTHENTICATOR_INTENT);
    List<ResolveInfo> resolveInfos =
        pm.queryIntentServices(authenticatorIntent, PackageManager.GET_META_DATA);
    if (resolveInfos == null) {
      Log.e(Constants.TAG, "Unable to query authenticator types");
      return Collections.emptyMap();
    }
    Map<String, String> authenticatorTypesToPackageNameMap = new HashMap<>();
    for (ResolveInfo resolveInfo : resolveInfos) {
      XmlResourceParser parser =
          resolveInfo.serviceInfo.loadXmlMetaData(pm, AccountManager.AUTHENTICATOR_META_DATA_NAME);
      if (parser != null) {
        try {
          String packageName = resolveInfo.serviceInfo.packageName;
          String accountType = parseAccountType(pm, parser, packageName);
          authenticatorTypesToPackageNameMap.put(accountType, packageName);
        } catch (IOException | XmlPullParserException e) {
          // Shouldn't happen, parser returned by system should be well formed. Log just in case.
          Log.e(Constants.TAG, "Unable to parse authenticator type", e);
        } catch (PackageManager.NameNotFoundException e) {
          Log.e(Constants.TAG, "Application does not exist", e);
        } catch (NoSuchFieldException | ClassNotFoundException | IllegalAccessException e) {
          Log.e(Constants.TAG, "Unable to get resource string value for accountType", e);
        }
      }
    }
    return authenticatorTypesToPackageNameMap;
  }

  private static String parseAccountType(
      PackageManager pm, XmlResourceParser parser, String packageName)
      throws IOException, XmlPullParserException, PackageManager.NameNotFoundException,
          NoSuchFieldException, ClassNotFoundException, IllegalAccessException {
    while (parser.next() != XmlPullParser.END_DOCUMENT) {
      if (parser.getEventType() != XmlPullParser.START_TAG
          || !parser.getName().equals(AccountManager.AUTHENTICATOR_ATTRIBUTES_NAME)) {
        continue;
      }
      Resources appResources = pm.getResourcesForApplication(packageName);
      AttributeSet attrs = Xml.asAttributeSet(parser);
      int accountTypeId = (int) getStyleableObject("AccountAuthenticator_accountType");
      return getResourceStringValue(appResources, attrs, accountTypeId);
    }
    return null;
  }

  /* android.R.styleable is not in the public API. However, we can still access its fields using
   * reflection. */
  private static String getResourceStringValue(Resources resources, AttributeSet set, int resId)
          throws NoSuchFieldException, ClassNotFoundException, IllegalAccessException {
    int[] attrs = (int[]) getStyleableObject("AccountAuthenticator");
    final TypedArray sa = resources.obtainAttributes(set, attrs);
    try {
      final TypedValue tv = sa.peekValue(resId);
      CharSequence data = null;
      if (tv != null && tv.type == TypedValue.TYPE_STRING) {
        if (tv.resourceId != 0) {
          data = resources.getString(tv.resourceId);
        } else {
          data = tv.string;
        }
      }
      return (data != null) ? data.toString() : null;
    } finally {
      sa.recycle();
    }
  }

  private static Object getStyleableObject(String name)
          throws ClassNotFoundException, NoSuchFieldException, IllegalAccessException {
    Field field = Class.forName("android.R$styleable").getDeclaredField(name);
    return field.get(null);
  }
}
